/*=========================================================================

Program:   Insight Segmentation & Registration Toolkit
Module:    $RCSfile: itkFMMPolynomial.h,v $
Language:  C++
Date:      $Date: 2004/08/06 12:38:33 $
Version:   $Revision: 1.35 $

Copyright (c) Insight Software Consortium. All rights reserved.
See ITKCopyright.txt or http://www.itk.org/HTML/Copyright.htm for details.

This software is distributed WITHOUT ANY WARRANTY; without even 
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __polynomial_h
#define __polynomial_h


#include <valarray>
#include <iostream>
#include "macros.h"

namespace fmi
{
   


/*
 This class represents a 3D polynomial. In this project we assume the
 following ordering of the polynomial:
 1, x, y, z, x^2, xy, xz, y^2, yz, z^2, x^3, x^2y, x^2z, ...
 These elements are stored in a single linear array, m_Data.

 For each degree this is equivalent to:
 
 Degree 0:
 1
 
 Degree 1:
 x
 y z

 Degree 2:
 xx
 xy xz
 yy yz zz

 Degree 3:
 xxx
 xxy xxz
 xyy xyz xzz
 yyy yyz yzz zzz

 Degree 4:
 xxxx
 xxxy xxxz
 xxyy xxyz xxzz
 xyyy xyyz xyzz xzzz
 yyyy yyyz yyzz yzzz zzzz

 Degree 5:
 xxxxx
 xxxxy xxxxz
 xxxyy xxxyz xxxzz
 xxyyy xxyyz xxyzz xxzzz
 xyyyy xyyyz xyyzz xyzzz xzzzz
 yyyyy yyyyz yyyzz yyzzz yzzzz zzzzz

 Degree 6:
 xxxxxx
 xxxxxy xxxxxz
 xxxxyy xxxxyz xxxxzz
 xxxyyy xxxyyz xxxyzz xxxzzz
 xxyyyy xxyyyz xxyyzz xxyzzz xxzzzz
 xyyyyy xyyyyz xyyyzz xyyzzz xyzzzz xzzzzz
 yyyyyy yyyyyz yyyyzz yyyzzz yyzzzz yzzzzz zzzzzz

 etc...

 Thus for each degree we have a lower triangular matrix describing the
 coefficients of the polynomial. Let d be the degree, then the ij'th element
 of this matrix (where i is the row and j is the column, both starting from
 zero) is

  d-i  i-j  j
 x    y    z

 For example, for d=3,i=2,j=1, we get x^(3-2) y^(2-1) x^(1) = xyz

 For a polynomial of degree d, the total number of terms is
 (d+1)*(2+(d+1)*(3+(d+1)))/6 

 The offset of the coefficient (d,i,j) from the beginning of the array
 is: d*(2+d*(3+d))/6 + i*(i+1)/2 + j
 

*/

class Polynomial
{

 private:


      unsigned int m_Size;
      int m_Degree;
      std::valarray< ScalarType > m_Data;


 public:

      static const unsigned int Offsets[];


      explicit Polynomial(int degree):
	 m_Size( (degree+1)*(2+(degree+1)*(3+(degree+1)))/6 ),
	 m_Degree( degree ),
	 m_Data( 0.0, m_Size ) // initialize to zero
      {
	 if(degree < -1)
	 {
	    std::cout << "error: degree too small" << std::endl;
	    abort();
	 }

      }
     Polynomial():
	 m_Size( 0 ),
	 m_Degree( -1 )
      {
      }

      Polynomial( const Polynomial& rhs ):
	 m_Size(rhs.m_Size),
	 m_Degree(rhs.m_Degree),
	 m_Data(rhs.m_Data)
      {
      }

      Polynomial& operator=( const Polynomial& rhs ) 
      {
	 if(this != &rhs)
	 {
	    // A valarray does not change size when using operator=
	    // so we must change it manually before calling it.
	    if(m_Size != rhs.GetSize())
	      m_Data.resize(rhs.GetSize());

	    m_Size = rhs.m_Size;
	    m_Degree = rhs.m_Degree;
	    m_Data = rhs.m_Data;
	 }
	 return *this;
      }
      ~Polynomial() {}


      int GetDegree() const { return m_Degree; }
      unsigned int GetSize() const { return m_Size; }

      void SetDegree(int degree)
      {
	 if(degree < -1)
	 {
	    std::cout << "error: degree too small" << std::endl;
	    abort();
	 }

	 m_Size = (degree+1)*(2+(degree+1)*(3+(degree+1)))/6;
	 m_Degree = degree;

	 m_Data.resize(m_Size);

	 //m_Data = 0.0;
	 this->fill_zero();

      }

      void SetDegreeNoFillZero(int degree)
      {

	 if(degree < -1)
	 {
	    std::cout << "error: degree too small" << std::endl;
	    abort();
	 }

	 m_Size = (degree+1)*(2+(degree+1)*(3+(degree+1)))/6;
	 m_Degree = degree;
	 m_Data.resize(m_Size);

      }


      
      // Return the coefficient of x^(d-i) * y^(i-j) * z^j
      ScalarType& operator() (const unsigned int d, const unsigned int i, 
                              const unsigned int j)
      {
	 return m_Data[GetOffset(d,i,j)];
      }
      ScalarType const & operator() (const unsigned int d, const unsigned int i, 
				     const unsigned int j) const 
      {
	 return m_Data[GetOffset(d,i,j)];
      }
      

      ScalarType& operator[] (const unsigned int i)
      {
	 return m_Data[ i ];
      }
      ScalarType const & operator[] (const unsigned int i) const
      {
	 return m_Data[ i ];
      }

      ScalarType& operator() (const unsigned int i)
      {
	 return m_Data[ i ];
      }
      ScalarType const & operator() (const unsigned int i) const
      {
	 return m_Data[ i ];
      }

      ScalarType* data_block()
      {
	 return &m_Data[0];
      }

      const ScalarType* data_block() const
      {
	 return &m_Data[0];
      }


      // Get the offset from the beginning of the array for a given
      // coefficient
      unsigned int GetOffset(const unsigned int d, 
			     const unsigned int i,
			     const unsigned int j) const 
      {
	 return d*(2+d*(3+d))/6 + i*(i+1)/2 + j;
      }
      
      // Same as previous function but assume j=0
      unsigned int GetOffset(const unsigned int d, 
			     const unsigned int i) const 
      {
	 return d*(2+d*(3+d))/6 + i*(i+1)/2;
      }

      // Same as previous function but assume i=0, j=0
      unsigned int GetOffset(const unsigned int d) const 
      {
	 return d*(2+d*(3+d))/6;
      }


      // Set the polynomial to zero
      void fill_zero()
      {
	 for ( unsigned int i=0; i < m_Size; ++i )
	    m_Data[i] = 0.0;
//	 m_Data = 0.0;
      }



      // Add a polynomial to this polynomial
      void Add( const Polynomial& poly)
      {
	 for ( unsigned int i=0; i < m_Size; ++i )
	    m_Data[i] += poly[i];
	 //m_Data += poly.m_Data;
      }




/*   
   // Evaluate the polynomial on a grid
   // the grid is passed as a linear array of values
   // where the x values changes the fastest and the 
   // z values change the slowest
   void EvaluateGrid(ScalarType* grid, const ScalarType* origin,
		       const ScalarType* spacing, const int* extent) const
     {

	ScalarType x,y,z;

	int xsize = extent[0];
	int ysize = extent[1];
	int zsize = extent[2];
	
	ScalarType dx = spacing[0];
	ScalarType dy = spacing[1];
	ScalarType dz = spacing[2];
	
	
	// Do the setup stage
	int c = 0;
	int i,j,k;
	
	for(k=0,z=origin[2]; k<zsize; k++,z+=dz)
	{
	   for(j=0,y=origin[1]; j<ysize; j++,y+=dy)
	   {
	      for(i=0,x=origin[0]; i<xsize; i++,x+=dx)
	      {
		 
//		 grid[c++];
		 
	      }
	   }
	}
	
	  
     }
*/
   
      void Print() const
      {
	 std::cout << "\n\n";

	 for(int m=0; m <= m_Degree; m++)
	 {
	    int begin = (m-1+1)*(2+(m-1+1)*(3+(m-1+1)))/6;
	    int end = (m+1)*(2+(m+1)*(3+(m+1)))/6;
	    std::cout << "degree: " << m << " begin: " << begin 
		      << " end: " << end << "\n";
	    for(int i=begin; i < end; i++)
	    {
	       std::cout << m_Data[i] << " ";
	    }
	    std::cout << "\n\n";
	 }
	 
	 
	 std::cout << "\n\n" << std::endl;
      }

};




}

#endif 

