#ifndef __solver_cxx
#define __solver_cxx

#include <time.h>
#include "solver.h"



// Include the iml implementation of gmres.
#include "gmres.h"

#include "macros.h"

#include <iostream>

namespace fmi
{



   solver::
   solver():
      m_PreconditioningFlag(false),
      m_MaxIterations(40),
      m_RestartIterations(40),
      m_Tolerance(1e-6),
      m_K(1),
      m_Tau(0.0),
      m_Lambda(0.0),
      m_MatrixWasModified(true),
      m_RightHandSideWasModified(true)
   {
      this->SetPolynomialDegree(-1);
   }

   

   void
   solver::
   SetPoints(const ScalarType* points, int numberOfPoints)
   {
      if(points == 0)
      {
	 std::cout << "Error: points is null." << std::endl;
	 abort();
      }


      m_Points.resize(3*numberOfPoints);

      for(int i=0;i<3*numberOfPoints;i++)
	 m_Points[i] = points[i]; 

      m_MatrixWasModified = true;
   }



   void
   solver::
   SetValues(const ScalarType* values, int numberOfPoints)
   {
      if(values == 0)
      {
	 std::cout << "Error: values is null." << std::endl;
	 abort();
      }


      m_Values.resize(numberOfPoints);

      for(int i=0;i<numberOfPoints;i++)
	 m_Values[i] = values[i]; 

      m_RightHandSideWasModified = true;
   }



   void 
   solver::
   SetPolynomialDegree(int d)
   {

      if(d < -1 || d > 2)
      {
	 std::cout << "Error: Invalid polynomial degree." << std::endl;
	 abort();
      }
	 

      m_PolynomialDegree = d;

      switch(m_PolynomialDegree)
      {
	 case -1:
	    m_NumberPolynomialTerms = 0;
	    break;
	 case 0:
	    m_NumberPolynomialTerms = 1;
	    break;
	 case 1:
	    m_NumberPolynomialTerms = 4;
	    break;
	 case 2:
	    m_NumberPolynomialTerms = 10;
	    break;
      }
      
      m_MatrixWasModified = true;
      m_RightHandSideWasModified = true;

   }	 
   


   void
   solver::
   Update()
   {

      if(m_Points.size() != 3*m_Values.size())
      {
	 std::cout << "Error. Mismatch in size of m_Points and m_Values."
		   << std::endl;
	 abort();
      }

      if(m_Points.size() <= 0 || m_Values.size() <= 0)
      {
	 std::cout << "Error: Number of points must be positive." << std::endl;
	 abort();
      }

      const ScalarType* sourceLandmarks = &m_Points[0]; 
      const ScalarType* values = &m_Values[0];
      const int numberOfPoints = m_Points.size()/3;


      if(m_MatrixWasModified)
      {

	 // allocate memory for the solution X
	 m_x.resize(numberOfPoints + m_NumberPolynomialTerms);

	 
	 // setup the M matrix of the  system: M * X = B.
	 m_A.SetupMatrix(sourceLandmarks,numberOfPoints,m_PolynomialDegree,
			 m_K,m_Tau,m_Lambda);
	 
	 
	 if(m_PreconditioningFlag)
	 {
	    std::cout << "Starting Preconditioning." << std::endl;

	    int start_time = clock();


	    int flag = pre.SetupPreconditioner(m_A,m_A.size1(),
					       sourceLandmarks,
					       numberOfPoints);

	    int end_time = clock();

	    std::cout << "Finished Preconditioning. Number Of Points: " 
		      << numberOfPoints << " Time taken: " << std::endl;

	    double elapsed_time = 
	       (end_time - start_time )/(double)CLOCKS_PER_SEC;
	    cout << elapsed_time       << " sec  = ";
	    cout << (elapsed_time/60.0) << " min = ";
	    cout << (elapsed_time/3600.0) << " hr" << endl;
	    

	    if(flag != 0)
	    {
	       m_PreconditioningFlag = false;
	       std::cout << "Warning: Not doing preconditioning." << std::endl;
	    }
	 }
	 
      }

      if(m_RightHandSideWasModified)
      {
	 // Setup the B matrix
	 int matrixRows = numberOfPoints + m_NumberPolynomialTerms;
	 m_b.resize(matrixRows);
   
	 int r=0;
	 for(;r<numberOfPoints;r++)
	 {
	    m_b[r] = values[r];
	 }

	 // Fill the remaining part of the matrix with zeros
	 for(;r<matrixRows;r++)
	 {
	    m_b[r] = 0.0;
	 }
      }
 

      H.resize(m_RestartIterations+1,m_RestartIterations+1);
      H = H * 0.0;

      // Set to zero
      m_x = m_x * 0.0;

      int maxit = m_MaxIterations;
      int resit = m_RestartIterations;
      ScalarType tol = m_Tolerance;
   
      
      PRINT(maxit);
      
      std::cout << "Starting GMRES." << std::endl;

      GMRES(m_A,m_x,m_b,pre,H,resit,maxit,tol);
   
      std::cout << "Finished GMRES." << std::endl;

      PRINT(maxit);

      //std::cout << "maxit: " << maxit << std::endl;

//      if(maxit >= m_MaxIterations)
      {
//	std::cout << "Warning: maximum number of iterations reached." 
//		  << std::endl;
	PRINTND(m_MaxIterations);
	PRINTND(maxit);
	PRINTND(m_Tolerance);
	PRINTND(tol);
      }
      
      m_MatrixWasModified = false;
      m_RightHandSideWasModified = false;

   }



}

#endif
